import random
import sys
sys.setrecursionlimit(5000)
from fractions import Fraction

def manche():
    """simulation du lancer de chaque dé avec victoire pour A si la somme est paire, pour B si la somme est impaire"""
    lancer_A = random.randint(1,6)
    lancer_B = random.randint(1,6)
    somme = lancer_A + lancer_B
    if somme % 2 == 0:
        return "A"
    else:
        return "B"

def test_equite(nb_manches):
    """test l'équité du jeu par simulation"""
    victoire_A = 0
    victoire_B = 0
    jeu = ""
    for i in range(nb_manches):
        jeu = manche()
        if jeu == "A":
            victoire_A = victoire_A + 1
        else:
            victoire_B = victoire_B + 1
    return victoire_A / nb_manches

def deroule_partie_2_1():
    """renvoie les possibilités de parties lorsque le score est à (2,1)"""
    lancer = manche() # 4e manche
    if lancer == "A":
        return "A" # le score est de (3,1) et A a gagné
    else: # B a gagné la 4e manche donc le score est à (2,2) et la partie continue
        lancer = manche() # 5e manche
        if lancer == "A": # A a gagné la 5e manche donc le score est à (3,2) et A a gagné la partie
            return "A"
        else: # B a gagné la 5e manche donc le score est à (2,3) et B a gagné la partie
            return "B"

def simulation_deroule_2_1(nb_parties):
    """on simule un grand nombre de parties pour connaitre la fréquence de victoire """
    victoire_A = 0 # penser à faire un copier/coller du test d'équité
    victoire_B = 0
    partie = ""
    for i in range(nb_parties):
        partie = deroule_partie_2_1()
        if partie == "A":
            victoire_A = victoire_A + 1
        else:
            victoire_B = victoire_B + 1
    return victoire_A / nb_parties




def deroule_partie_2_0():
    """renvoie les possibilités de parties lorsque le score est à (2,0)"""
    lancer = manche() # 4e manche
    if lancer == "A":
        return "A" # le score est de (3,0) et A a gagné
    else: # B a gagné la 3e manche donc le score est à (2,1) et on peut reprendre le deroule_partie_2_1
        return deroule_partie_2_1()


def simulation_deroule_2_0(nb_parties):
    """on simule un grand nombre de parties pour connaitre la fréquence de victoire """
    victoire_A = 0 # penser à faire un copier/coller du test d'équité
    victoire_B = 0
    partie = ""
    for i in range(nb_parties):
        partie = deroule_partie_2_0()
        if partie == "A":
            victoire_A = victoire_A + 1
        else:
            victoire_B = victoire_B + 1
    return victoire_A / nb_parties

def deroule_partie_1_0():
    """renvoie les possibilités de parties lorsque le score est à (1,0)"""
    lancer = manche() # 2e manche
    if lancer == "A": # A a gagne la 2e manche et le score est à (2,0) et on peut reprendre le déroulé
        return deroule_partie_2_0()
    else: # B a gagné la 2e manche donc le score est à (1,1) et la suite est équitable chacun a autant de chance de gagner donc on choisit le vainqueur au hasard
        return random.choice(["A","B"])

def simulation_deroule_1_0(nb_parties):
    """on simule un grand nombre de parties pour connaitre la fréquence de victoire """
    victoire_A = 0 # penser à faire un copier/coller du test d'équité
    victoire_B = 0
    partie = ""
    for i in range(nb_parties):
        partie = deroule_partie_1_0()
        if partie == "A":
            victoire_A = victoire_A + 1
        else:
            victoire_B = victoire_B + 1
    return victoire_A / nb_parties


def partis(cible,a,b):
    """renvoie la probabilité de victoire de A à partir de l'état [a,b] dans le problème historique des partis"""
    if a == cible:
        return 1
    elif b == cible:
        return 0
    else:
        return 0.5 * partis(cible, a+1, b) + 0.5 * partis(cible, a, b+1)

def montant_partis(pot,cible,a,b):
    """renvoie la répartition de la somme en cas d'arrêt du jeu à l'état [a,b] pour un nombre cible de victoires à atteindre"""
    montant_a = pot * partis(cible,a,b)
    montant_b = pot - montant_a
    return (montant_a, montant_b)

def partis_memo(cible,a,b):
    """calcule le tableau de probabilité à tous les états de jeu possibles et renvoie la probabilite demandée pour A"""
    cache = [[ 0 for l in range (cible)] + [0] for k in range(cible)] + [[1 for k in range(cible)]+[""]]
    for ligne in range(cible-1,-1,-1):
        for colonne in range(cible-1,-1,-1):
            if ligne == colonne:
                cache[ligne][colonne]= 0.5
            else:
                cache[ligne][colonne] =  0.5*cache[ligne+1][colonne] + 0.5*cache[ligne][colonne+1]
    return Fraction(cache[a][b])

def pos_diag(n):
    """détermine les listes des positions de chaque fausse diagonale"""
    colonne = n - 1
    ligne = n -1
    rang = 1
    liste = [[]]
    while ligne >= 0:
        if colonne >= 0:
            indice = (n, colonne)
            pos = [indice]
            for i in range(n-colonne):
                nouveau = (indice[0] - 1, indice[1] + 1)
                pos = pos + [nouveau]
                indice  = nouveau
            colonne = colonne - 1
        else:
            indice = (ligne, 0)
            pos = [indice]
            for i in range(ligne):
                nouveau = (indice[0] - 1, indice[1] + 1)
                pos = pos + [nouveau]
                indice  = nouveau
            ligne = ligne - 1
        liste = liste + [pos]
    return liste

def partis_diag(n,a,b):
    """calcule uniquement la diagonale courante jusqu'à atteindre celle contenant l'état [a,b] dans le cas d'une partie à n victoires"""
    colonne = n - 1
    ligne = n - 1
    dico = {(n,n-1):1, (n-1,n):0}
    liste_pos = []
    while (a,b) not in liste_pos and ligne >= 0:
        if colonne >= 0:
            indice = (n, colonne)
            liste_pos = [indice]
            for i in range(n-colonne):
                nouveau = (indice[0] - 1, indice[1] + 1)
                liste_pos = liste_pos + [nouveau]
                indice  = nouveau
            colonne = colonne - 1
        else:
            indice = (ligne, 0)
            liste_pos = [indice]
            for i in range(ligne):
                nouveau = (indice[0] - 1, indice[1] + 1)
                liste_pos = liste_pos + [nouveau]
                indice  = nouveau
            ligne = ligne - 1
        dico_ancien = dico
        dico = {}
        for pos in liste_pos:
            if pos[0] == n:
                dico[pos] = 1
            elif pos[1] == n:
                dico[pos] = 0
            else:
                dico[pos] = 0.5*dico_ancien[(pos[0] + 1, pos[1])] +0.5*dico_ancien[(pos[0], pos[1] + 1)]
    return dico[(a,b)]

def deroule_partie_1_3():
    """renvoie les possibilités de parties lorsqu'il manque une victoire à A et 3 victoires à B"""
    lancer = manche() # nouvelle manche
    if lancer == "A":
        return "A" # A a gagné la manche qu'il lui manquait et il a gagné la partie
    else: # B a gagné la manche et il lui manque 2 victoires (et 1 pour A)
        lancer = manche()
        if lancer == "A":# A a gagné la manche qu'il lui manquait et il a gagné la partie
            return "A"
        else: # B a gagné la manche et il lui manque 1 victoire (et 1 pour A)
            lancer = manche()
            if lancer == "A":# A a gagné la manche qu'il lui manquait et il a gagné la partie
                return "A"
            else:# B a gagné la manche et il a gagné la partie
                return "B"





def simulation_deroule_1_3(nb_parties):
    """on simule un grand nombre de parties pour connaitre la fréquence de victoire """
    victoire_A = 0 # penser à faire un copier/coller du test d'équité
    victoire_B = 0
    partie = ""
    for i in range(nb_parties):
        partie = deroule_partie_1_3()
        if partie == "A":
            victoire_A = victoire_A + 1
        else:
            victoire_B = victoire_B + 1
    return victoire_A / nb_parties

def lissage_1_3(nb_parties, nb_echantillons):
    somme = 0
    for k in range(nb_echantillons):
        serie = simulation_deroule_1_3(nb_parties)
        somme = somme + serie
    return somme / nb_echantillons

def comparatif_solutions(cible,a,b):
    """on teste les réponses formulées par chaque personnage historique"""
    dico = {}
    if a + b >= cible:
        diff = a + b - cible
        dico["X"] = (Fraction(2*a-diff, 2*cible), Fraction(2*b-diff,2*cible))
    else:
        dico["X"] = None

    dico["Y"] = (Fraction(a-b+cible,2*cible),Fraction(cible-a+b,2*cible))
    dico["Luca"] = (Fraction(a,a+b),Fraction(b,a+b))
    dico["Nicolas"] = (Fraction(cible+(a-b), 2*cible), Fraction(cible-(a-b),2*cible))
    dico["Lorenzo"] = (Fraction(a+2*cible-1-b,2*(2*cible-1)), Fraction(b+2*cible-1-a,2*(2*cible-1)))
    total_jerome = ((cible - a)*(cible-a+1) + (cible-b)*(cible-b+1) )// 2
    dico["Jerome"] = (Fraction(((cible - b)*(cible-b+1))//2, total_jerome), Fraction(((cible - a)*(cible-a+1))//2, total_jerome))
    dico["Blaise"] = (partis_memo(cible,a,b), Fraction(1-partis_memo(cible,a,b)))
    dico["Pierre"] = (partis_memo(cible,a,b), Fraction(1-partis_memo(cible,a,b)))
    for cle in dico:
        print(f"La solution de {cle} est la répartition {dico[cle]}.")

